#include <stdlib.h>      


#define order 200
#define numberoftasks 4
#define grainsize 1000
#define partitions 8      

float crandom( int *seed);
void SubA( float a[order][order],
           float x[order],
           float b[order],
           int *n,
           int *lockid,
           int *subset,
           int *outstanding,
           int *parts,
           int first[],
           int locks[],
           int *taskindex,
           int *tasks,
           float work[],
           float residual[],
           float *tolerance,
           int *done,
           int event[],
           int *activeevent);


main( )
{
/*
;
;  This is a main function for the example in section 7.3.
;  This example, in parallel, solves a system of linear equations in
;  the form [A]{X}={B} by the asynchronous Gauss-Seidal Iteration.
;  
*/
   static int parts,
              n,
              lockid,
              event[2],
              subset,
              locks[partitions],
              outstanding,
              tasks,
              first[partitions+1] = 
                   { 0, 25, 50, 75, 100, 125, 150, 175, 200 },
              j4temp,
              k4temp,
              seed = 88888888,
              taskid[numberoftasks],
              taskindex[numberoftasks],
              activeevent,
              done;
   static float a[order][order],
                x[order],
                b[order],
                work[numberoftasks],
                residual[numberoftasks],
                tolerance = 0.001;

/*  generate data in a sequential procedure  */

   printf(">>> Generate data ...\n");
   tasks = numberoftasks;
   n = order;
   parts = partitions;
   for ( j4temp = 0; j4temp < n; j4temp ++ )
         for (k4temp = 0; k4temp < n; k4temp ++ )
              a[k4temp][j4temp] = crandom( &seed );
   for ( j4temp = 0; j4temp < n; j4temp ++ )
   {
      for ( k4temp = 0; k4temp < n; k4temp ++ )
            a[j4temp][j4temp] = a[j4temp][j4temp] + a[j4temp][k4temp];
      a[j4temp][j4temp] = a[j4temp][j4temp] * 5.0;
   }
   for ( j4temp = 0; j4temp < n; j4temp ++ )
   {
      b[j4temp] = 0.0;
      for ( k4temp = 0; k4temp < n; k4temp ++ )
            b[j4temp] = b[j4temp] + a[j4temp][k4temp];
   }
   for ( j4temp = 0; j4temp < n; j4temp ++ )
         x[j4temp] = 0.0;

/*  create parallel locks  */

   CreateParallelLock(&lockid);
   for ( j4temp = 0; j4temp < parts; j4temp ++ )
         CreateParallelLock(&locks[j4temp]);

/*  create parallel events  */

   CreateParallelEvent(&event[0]);
   CreateParallelEvent(&event[1]);
   activeevent = 0;

/*  determine length of subsets  */

   subset = __min( grainsize / ( 2 * n + 5), order / numberoftasks);
   if(subset == 0) subset = 1;

/*  initialize  */

   outstanding = 0;
   InitializeParallelEvent(&event[0], &tasks);

/*  solve in parallel  */

   printf(">>> Solve in parallel ...\n");
   for ( j4temp = 0; j4temp < tasks - 1; j4temp ++ )
   {
      taskindex[j4temp] = j4temp;
      Dispatch_18(&taskid[j4temp], SubA, a,
                                         x,
                                         b,
                                         &n,
                                         &lockid,
                                         &subset,
                                         &outstanding,
                                         &parts,
                                         first,
                                         locks,
                                         &taskindex[j4temp],
                                         &tasks,
                                         work,
                                         residual,
                                         &tolerance,
                                         &done,
                                         event,
                                         &activeevent);
   }
   taskindex[tasks-1] = tasks - 1;
   SubA(a, x, b, &n, &lockid, &subset, &outstanding, &parts,
        first, locks, &taskindex[tasks-1], &tasks, work, residual,
        &tolerance, &done, event, &activeevent);

/*  terminate tasks  */

   WaitForAndTerminateAllTasks( );

/*  delete parallel lock  */

   DeleteParallelLock(&lockid);
   for ( j4temp = 0; j4temp < parts; j4temp ++ )
         DeleteParallelLock(&locks[j4temp]);

/*  delete parallel events  */

   DeleteParallelEvent(&event[0]);
   DeleteParallelEvent(&event[1]);

/*  output  */

   printf("     x[  0]= %2.4f\n", x[  0]);
   printf("     x[ 49]= %2.4f\n", x[ 49]);
   printf("     x[ 99]= %2.4f\n", x[ 99]);
   printf("     x[149]= %2.4f\n", x[149]);
   printf("     x[199]= %2.4f\n", x[199]);
}

void SubA( float a[order][order],
           float x[order],
           float b[order],
           int *n,
           int *lockid,
           int *subset,
           int *outstanding,
           int *parts,
           int first[],
           int locks[],
           int *taskindex,
           int *tasks,
           float work[],
           float residual[],
           float *tolerance,
           int *done,
           int event[],
           int *activeevent)
{
/*
;
;  function to solve a system of liner equations by asynchronous Gauss-Seidal
;  iteration
;  (A)Parameters:
;     1.a: left side positive definite matrix, dimension[n][n]
;     2.x: solution to be determined, dimension[n]
;     3.b: right side vector, dimension[n]
;     4.n: order
;     5.lockid: identifier of lock that protects distribution of subsets
;     6.subset: length of subset
;     7.outstanding: outstanding element that must be initialized to zero
;     8.parts: number of critical subsections
;     9.first: the first index in each critical subsection,
;              dimension[parts+1],
;              in which the last element = n
;    10.locks: identifier of locks for each critical subsection,
;              dimension[parts]
;    11.taskindex: task ordinal number
;    12.tasks: number of tasks
;    13.work: working array for computing an increment, dimension[tasks]
;    14.residual: working array for the maximal residual resulted in
;                 each task, dimension[tasks]
;    15.tolerance: tolerance of residual
;    16.done: working variable to indicate if the job has been done
;    17.event: identifiers of parallel event that handles task
;              synchronization, dimension[2]
;              The first parallel event, event[0], must be initialized
;              in the caller.
;    18.activeevent: the active parallel event which must be initialized
;                    to zero in the caller

/*  private variables  */

   int j4temp$, k4temp$, l4temp$, m4temp$, n4temp$, yes$;

/*  entry point to a loop  */

   *done = 0;
label_10:    
   residual[*taskindex] = 0.0;
   l4temp$ = 0;
/*
;
;  [critical section]
;  distribute subsets
;
*/
label_20:    
   LockonParallelLock(lockid);
   if( *outstanding > *n - 1 )
   {
      UnlockParallelLock(lockid);
      goto label_30;
   }
   j4temp$ = *outstanding;
   *outstanding = *outstanding + *subset;
   UnlockParallelLock(lockid);
/*
;
;  [parallelism]
;  set initial value for the increment
;  total size = 2 * n + 5
;
*/
   for ( k4temp$ = j4temp$;
         k4temp$ < __min( j4temp$ + *subset, *n);
         k4temp$ ++ )
   {
         work[*taskindex] = b[k4temp$];
/*
;
;  [implicit parallelism]
;  compute in each critical subsection
;
*/
      for ( m4temp$ = 0; m4temp$ < *parts; m4temp$ ++ )
      {
         LockonParallelLock(&locks[m4temp$]);
         for ( n4temp$ = first[m4temp$];
               n4temp$ < first[m4temp$+1];
               n4temp$ ++ )
               work[*taskindex] = work[*taskindex] - a[k4temp$][n4temp$] *
                                                     x[n4temp$];
         UnlockParallelLock(&locks[m4temp$]);
      }
/*
;
;  [parallelism]
;  compute an increment
;
*/
      work[*taskindex] = work[*taskindex] / a[k4temp$][k4temp$];
/*
;
;  [implicit parallelism]
;  update an element of {x}
;
*/
      while ( first[l4temp$+1] <= k4temp$)
         l4temp$ = l4temp$ +1;
      LockonParallelLock(&locks[l4temp$]);
      x[k4temp$] = x[k4temp$] + work[*taskindex];
      UnlockParallelLock(&locks[l4temp$]);
/*
;
;  [parallelism]
;  residual
;
*/
      if( work[*taskindex] < 0.0 ) work[*taskindex] = - work[*taskindex];
      if( work[*taskindex] > residual[*taskindex] )
          residual[*taskindex] = work[*taskindex];
   }
   goto label_20;
/*
;
;  [task synchronization]
;  complete a cycle of parallel event
;
*/
label_30:    
   PostParallelEvent(&event[*activeevent]);
   IfCompleteParallelEvent(&event[*activeevent], &yes$);
/*
;
;  [subsequently sequential fragment]
;  check accuracy, and initialize outstanding element and
;  another parallel event
;
*/
   if( yes$ == 0)
   {
      k4temp$ = 2;
      for ( j4temp$ = 0; j4temp$ < *tasks; j4temp$ ++ )
      {
         if( residual[j4temp$] > *tolerance)
         {
            *outstanding = 0;
            switch (*activeevent)
            {
               case 0:
                  *activeevent = 1;
                  k4temp$ = 0;
                  break;
               case 1:
                  *activeevent = 0;
                  k4temp$ = 1;
                  break;
            }
            InitializeParallelEvent(&event[*activeevent], tasks);
            break;
         }
      }
      if( k4temp$ == 2)
      {
         *done = 1;
         CompleteParallelEvent(&event[*activeevent]);
      }
      else
         CompleteParallelEvent(&event[k4temp$]);
   }

/*  return for convergence of solution  */

   if( *done == 1) return;

/*  go to one more loop  */

   goto label_10;
}
#define m 2147483647
#define a 16807
float crandom(int *seed)
{
   double rtempa, rtempb;
   float rtemp;
   int itemp;
   rtempa = (double) *seed * a;
   itemp = (int) (rtempa / (double) m);
   rtempb = (double) m * (double) itemp;
   *seed = (int) (rtempa - rtempb);
   rtemp = (double) *seed / (double) m;
   return rtemp;
}
